/* ************************************************************************
*   File: class.c                                       Part of CircleMUD *
*  Usage: Source file for class-specific code                             *
*                                                                         *
*  All rights reserved.  See license.doc for complete information.        *
*                                                                         *
*  Copyright (C) 1993, 94 by the Trustees of the Johns Hopkins University *
*  CircleMUD is based on DikuMUD, Copyright (C) 1990, 1991.               *
************************************************************************ */

/*
 * This file attempts to concentrate most of the code which must be changed
 * in order for new classes to be added.  If you're adding a new class,
 * you should go through this entire file from beginning to end and add
 * the appropriate new special cases for your new class.
 */

#include "conf.h"
#include "sysdep.h"
#include "math.h"

#include "structs.h"
#include "db.h"
#include "utils.h"
#include "spells.h"
#include "interpreter.h"
#include "constants.h"

/* local functions */
void snoop_check(struct char_data *ch);
int parse_class(char arg);
bitvector_t find_class_bitvector(const char *arg);
byte saving_throws(int class_num, int type, int level);
int thaco(int class_num, int level);
void roll_real_abils(struct char_data *ch);
void do_start(struct char_data *ch);
int backstab_mult(int level);
int invalid_class(struct char_data *ch, struct obj_data *obj);
int level_exp(int chclass, int level, int race);
int saving_throw_formula(int high_val, int low_val, int level);

/* Names first */

const char *class_abbrevs[] = {
  "Apr",
  "Swf",
  "Swm",
  "Whb",
  "Mag",
  "Arc",
  "Wiz",
  "Ran",
  "Ave",
  "Dru",
  "Cle",
  "Mon",
  "Pri",
  "Thi",
  "Rog",
  "Ass",
  "Cru",
  "Pal",
  "Tem",
  "Ele",
  "Ill",
  "Sor",
  "War",
  "Lic",
  "Dre",
  "\n"
};

const char *pc_class_types[] = {
  "Apprentice",
  "Sword fighter",
  "Sword master",
  "Whirlblade",
  "Mage",
  "Archmage",
  "Wizard",
  "Ranger",
  "Avenger",
  "Druid",
  "Cleric",
  "Monk",
  "Priest",
  "Thief",
  "Rogue",
  "Assassin",
  "Crusader",
  "Paladin",
  "Templar",
  "Elementalist",
  "Illusionist",
  "Sorcerer",
  "Warlock",
  "Lich",
  "Dreadlord",
  "\n"
};

/*

swordfighter, mage,     ranger,  cleric, thief,    crusader,     elementalist,   warlock,

swordmaster,  archmage, avenger, monk,   rogue,    paladin,      illusionist,    lich
                            
whirlblade,   wizard,   druid,   priest, assassin, templar,      sorcerer,       dreadlord

*/

/*
 * The code to interpret a class letter -- used in interpreter.c when a
 * new character is selecting a class and by 'set class' in act.wizard.c.
 */

int parse_class(char arg)
{
  arg = LOWER(arg);  

  switch (arg) {
	case 'a': return CLASS_APPRENTICE; 
    case 'b': return CLASS_SWORDFIGHTER;
    case 'c': return CLASS_SWORDMASTER;
    case 'd': return CLASS_WHIRLBLADE;
    case 'e': return CLASS_MAGE;
    case 'f': return CLASS_ARCHMAGE;
    case 'g': return CLASS_WIZARD;
    case 'h': return CLASS_RANGER;
    case 'i': return CLASS_AVENGER;
    case 'j': return CLASS_DRUID;
	case 'k': return CLASS_CLERIC;
    case 'l': return CLASS_MONK;
    case 'm': return CLASS_PRIEST;
    case 'n': return CLASS_THIEF;
    case 'o': return CLASS_ROGUE;
	case 'p': return CLASS_ASSASIN;
    case 'q': return CLASS_CRUSADER;
    case 'r': return CLASS_PALADIN;
    case 's': return CLASS_TEMPLAR;
    case 't': return CLASS_ELEMENTALIST;
	case 'u': return CLASS_ILLUSIONIST;
	case 'v': return CLASS_SORCERER;
	case 'w': return CLASS_WARLOCK;
	case 'x': return CLASS_LICH;
	case 'y': return CLASS_DREADLORD;
	default : return CLASS_UNDEFINED;
  }
}

/*
 * bitvectors (i.e., powers of two) for each class, mainly for use in
 * do_who and do_users.  Add new classes at the end so that all classes
 * use sequential powers of two (1 << 0, 1 << 1, 1 << 2, 1 << 3, 1 << 4,
 * 1 << 5, etc.) up to the limit of your bitvector_t, typically 0-31.
 */
bitvector_t find_class_bitvector(const char *arg)
{
  size_t rpos, ret = 0;

  for (rpos = 0; rpos < strlen(arg); rpos++)
    ret |= (1 << parse_class(arg[rpos]));

  return (ret);
}


/*
 * These are definitions which control the guildmasters for each class.
 *
 * The first field (top line) controls the highest percentage skill level
 * a character of the class is allowed to attain in any skill.  (After
 * this level, attempts to practice will say "You are already learned in
 * this area."
 * 
 * The second line controls the maximum percent gain in learnedness a
 * character is allowed per practice -- in other words, if the random
 * die throw comes out higher than this number, the gain will only be
 * this number instead.
 *
 * The third line controls the minimu percent gain in learnedness a
 * character is allowed per practice -- in other words, if the random
 * die throw comes out below this number, the gain will be set up to
 * this number.
 * 
 * The fourth line simply sets whether the character knows 'spells'
 * or 'skills'.  This does not affect anything except the message given
 * to the character when trying to practice (i.e. "You know of the
 * following spells" vs. "You know of the following skills"
 */

#define SPELL	0
#define SKILL	1

/* #define LEARNED_LEVEL	0  % known which is considered "learned" */
/* #define MAX_PER_PRAC		1  max percent gain in skill per practice */
/* #define MIN_PER_PRAC		2  min percent gain in skill per practice */
/* #define PRAC_TYPE		3  should it say 'spell' or 'skill'?	*/

int prac_params[4][NUM_CLASSES] = {
  /* MAG	CLE	THE	WAR */
  { 95,		95,	85,	80	},	/* learned level */
  { 100,	100,	12,	12	},	/* max per practice */
  { 25,		25,	0,	0	},	/* min per practice */
  { SPELL,	SPELL,	SKILL,	SKILL	},	/* prac name */
};


/*
 * ...And the appropriate rooms for each guildmaster/guildguard; controls
 * which types of people the various guildguards let through.  i.e., the
 * first line shows that from room 3017, only MAGIC_USERS are allowed
 * to go south.
 *
 * Don't forget to visit spec_assign.c if you create any new mobiles that
 * should be a guild master or guard so they can act appropriately. If you
 * "recycle" the existing mobs that are used in other guilds for your new
 * guild, then you don't have to change that file, only here.
 *
struct guild_info_type guild_info[] = {


  { CLASS_MAGIC_USER,	3017,	SCMD_SOUTH	},
  { CLASS_CLERIC,	3004,	SCMD_NORTH	},
  { CLASS_THIEF,	3027,	SCMD_EAST	},
  { CLASS_WARRIOR,	3021,	SCMD_EAST	},


  { -999 ,	5065,	SCMD_WEST	},

  { -1, NOWHERE, -1}
};
*/


/*
 * Saving throws for:
 * MCTW
 *   PARA, ROD, PETRI, BREATH, SPELL
 *     Levels 0-40
 *
 * Do not forget to change extern declaration in magic.c if you add to this.
 */

byte saving_throws(int class_num, int type, int level)
{
  int high_val, low_val;

  switch(class_num) {
  case CLASS_MAGE:
  case CLASS_ARCHMAGE:
  case CLASS_WIZARD:
  case CLASS_ELEMENTALIST:
  case CLASS_ILLUSIONIST:
  case CLASS_SORCERER:
    switch(type) {
    case SAVING_PARA:   high_val = 71; low_val = 29; break;
    case SAVING_ROD:    high_val = 54; low_val = 8; break;
    case SAVING_PETRI:  high_val = 65; low_val = 14; break;
    case SAVING_BREATH: high_val = 75; low_val = 24; break;
    case SAVING_SPELL:  high_val = 60; low_val = 9; break;
    default: log("SYSERR: Invalid saving throw type."); return 99;
	}
    break;
  case CLASS_CLERIC:
  case CLASS_MONK:
  case CLASS_PRIEST:
  case CLASS_CRUSADER:
  case CLASS_PALADIN:
  case CLASS_TEMPLAR:
    switch(type) {
    case SAVING_PARA:   high_val = 58; low_val = 5; break;
    case SAVING_ROD:    high_val = 71; low_val = 27; break;
    case SAVING_PETRI:  high_val = 65; low_val = 22; break;
    case SAVING_BREATH: high_val = 80; low_val = 37; break;
    case SAVING_SPELL:  high_val = 75; low_val = 32; break;
    default: log("SYSERR: Invalid saving throw type."); return 99;
    }
    break;
  case CLASS_THIEF:
  case CLASS_ROGUE:
  case CLASS_ASSASIN: 
  case CLASS_RANGER:
  case CLASS_AVENGER:
  case CLASS_DRUID:
    switch(type) {
    case SAVING_PARA:   high_val = 65; low_val = 36; break;
    case SAVING_ROD:    high_val = 70; low_val = 13; break;
    case SAVING_PETRI:  high_val = 60; low_val = 31; break;
    case SAVING_BREATH: high_val = 80; low_val = 51; break;
    case SAVING_SPELL:  high_val = 75; low_val = 17; break;
    default: log("SYSERR: Invalid saving throw type."); return 99;
    }
    break;
  case CLASS_SWORDFIGHTER:
  case CLASS_WHIRLBLADE:
  case CLASS_SWORDMASTER:
    switch(type) {
    case SAVING_PARA:   high_val = 70; low_val = 8; break;
    case SAVING_ROD:    high_val = 80; low_val = 16; break;
    case SAVING_PETRI:  high_val = 75; low_val = 13; break;
    case SAVING_BREATH: high_val = 85; low_val = 13; break;
    case SAVING_SPELL:  high_val = 85; low_val = 21; break;
    default: log("SYSERR: Invalid saving throw type."); return 99;
    }
    break;
  case CLASS_WARLOCK:
  case CLASS_LICH:
  case CLASS_DREADLORD:
    switch(type) {
    case SAVING_PARA:   high_val = 60; low_val = 3; break;
    case SAVING_ROD:    high_val = 80; low_val = 16; break;
    case SAVING_PETRI:  high_val = 75; low_val = 13; break;
    case SAVING_BREATH: high_val = 50; low_val = 1; break;
    case SAVING_SPELL:  high_val = 85; low_val = 17; break;
    default: log("SYSERR: Invalid saving throw type."); return 99;
    }
    break;
	
	default:
    log("SYSERR: Invalid class passed to saving throw.");
    return 99;
  }
  if (level == 0)
    return 100;
  else if(level >= LVL_IMMORT)
    return 0;
  else
	return(saving_throw_formula(high_val, low_val, level));
    
  /* Should not get here unless something is wrong. */
  return 100;
}

int saving_throw_formula(int high_val, int low_val, int level)
{
  float x = 0;

  x += high_val;
  x -= low_val;
  x /= (LVL_IMMORT - 1);
  x *= (level - 1);
  x *= -1;
  x += high_val;
  return x;
}

/* THAC0 for classes and levels.  (To Hit Armor Class 0) */
int thaco(int class_num, int level)
{
  switch (class_num) {
  case CLASS_MAGE:
    if (level == 0)
      return 100;
    else if (level >= LVL_IMMORT)
      return 0;
    else 
      return (20 - ((10 * level) / (LVL_IMMORT - 1)));
  case CLASS_CLERIC:
    if (level == 0)
      return 100;
    else if (level >= LVL_IMMORT)
      return 0;
    else 
      return (20 - ((18 * level) / (LVL_IMMORT - 1)));
  case CLASS_THIEF:
    if (level == 0)
      return 100;
    else if (level >= LVL_IMMORT)
      return 0;
    else 
      return (20 - ((16 * level) / (LVL_IMMORT - 1)));
  case CLASS_SWORDFIGHTER:
    if (level == 0)
      return 100;
    else if (level >= LVL_IMMORT)
      return 0;
    else 
	  return (((20 - ((28 * level) / (LVL_IMMORT - 1))) <= 1) ? 1 : (20 - ((28 * level) / (LVL_IMMORT - 1))));
  default:
    log("SYSERR: Unknown class in thac0 chart.");
  }

  /* Will not get there unless something is wrong. */
  return 100;
}


/*
 * Roll the 6 stats for a character... each stat is made of the sum of
 * the best 3 out of 4 rolls of a 6-sided die.  Each class then decides
 * which priority will be given for the best to worst stats.
 */
void roll_real_abils(struct char_data *ch)
{
  int i, j, k, temp;
  ubyte table[6];
  ubyte rolls[4];

  for (i = 0; i < 6; i++)
    table[i] = 0;

  for (i = 0; i < 6; i++) {

    for (j = 0; j < 4; j++)
      rolls[j] = rand_number(1, 6);

    temp = rolls[0] + rolls[1] + rolls[2] + rolls[3] -
      MIN(rolls[0], MIN(rolls[1], MIN(rolls[2], rolls[3])));

    for (k = 0; k < 6; k++)
      if (table[k] < temp) {
	temp ^= table[k];
	table[k] ^= temp;
	temp ^= table[k];
      }
  }
  
  
  switch (GET_RACE(ch)) {
  case RACE_HOBBIT_HARFOOT:
    ch->real_abils.cha = table[0];
	ch->real_abils.dex = table[1];
    ch->real_abils.intel = table[2];
    ch->real_abils.wis = table[3];
    ch->real_abils.con = table[4];
	ch->real_abils.str = table[5];
	ch->real_abils.cha += 1;
    break;
  case RACE_HOBBIT_STOOR:
	ch->real_abils.dex = table[0];
    ch->real_abils.cha = table[1];
    ch->real_abils.intel = table[2];
    ch->real_abils.wis = table[3];
    ch->real_abils.con = table[4];
	ch->real_abils.str = table[5];
	ch->real_abils.wis -= 1;
    break;
  case RACE_HOBBIT_FALLOHIDE:
	ch->real_abils.dex = table[0];
    ch->real_abils.cha = table[1];
    ch->real_abils.intel = table[2];
    ch->real_abils.con = table[3];
	ch->real_abils.wis = table[4];
    ch->real_abils.str = table[5];
    break;
  case RACE_ELF_VANYAR:
	ch->real_abils.intel = table[0];
	ch->real_abils.dex = table[1];
    ch->real_abils.wis = table[2];
    ch->real_abils.con = table[3];
	ch->real_abils.str = table[4];
    ch->real_abils.cha = table[5];
    ch->real_abils.str += 1;
	break;
  case RACE_ELF_NOLDOR:
	ch->real_abils.intel = table[0];
	ch->real_abils.wis = table[1];
	ch->real_abils.dex = table[2];
    ch->real_abils.cha = table[3];
    ch->real_abils.str = table[4];
	ch->real_abils.con = table[5];
	ch->real_abils.con += 1;
	ch->real_abils.str += 1;
    break;
  case RACE_ELF_TELERI:
	ch->real_abils.intel = table[0];
	ch->real_abils.con = table[1];
	ch->real_abils.dex = table[2];
	ch->real_abils.str = table[3];
	ch->real_abils.wis = table[4];
	ch->real_abils.cha = table[5];
    break;
  case RACE_HUMAN_AIEL:
	ch->real_abils.dex = table[0];
    ch->real_abils.con = table[1];
	ch->real_abils.str = table[2];
	ch->real_abils.intel = table[3];
	ch->real_abils.wis = table[4];
	ch->real_abils.cha = table[5];
	ch->real_abils.cha += 1;
    break;
  case RACE_HUMAN_NUMERON:
    ch->real_abils.str = table[0];
    ch->real_abils.wis = table[1];
	ch->real_abils.con = table[2];
	ch->real_abils.intel = table[3];
	ch->real_abils.dex = table[4];
	ch->real_abils.cha = table[5];
    ch->real_abils.cha += 2;
	ch->real_abils.str -= 1;
    break;
  case RACE_HUMAN_HILDOR:
    ch->real_abils.dex = table[0];
	ch->real_abils.intel = table[1];
	ch->real_abils.con = table[2];
	ch->real_abils.str = table[3];
    ch->real_abils.wis = table[4];
	ch->real_abils.cha = table[5];
    ch->real_abils.cha += 1;
    break;
  case RACE_HUMAN_ENGWAR:
	ch->real_abils.dex = table[0];
	ch->real_abils.con = table[1];
	ch->real_abils.str = table[2];
    ch->real_abils.intel = table[3];
	ch->real_abils.wis = table[4];
	ch->real_abils.cha = table[5];
    break;
  case RACE_ORC_URUKHAI:
	ch->real_abils.str = table[0];
    ch->real_abils.con = table[1];
	ch->real_abils.dex = table[2];
    ch->real_abils.wis = table[3];
	ch->real_abils.intel = table[4];
	ch->real_abils.cha = table[5];
    ch->real_abils.cha -= 1;
	ch->real_abils.con += 3;
	ch->real_abils.intel -= 2;
	ch->real_abils.wis -= 2;
    break;
  case RACE_ORC_OLEGHAI:
    ch->real_abils.str = table[0];
    ch->real_abils.con = table[1];
	ch->real_abils.dex = table[2];
	ch->real_abils.intel = table[3];
    ch->real_abils.wis = table[4];
	ch->real_abils.cha = table[5];
    ch->real_abils.cha -= 1;
	ch->real_abils.con += 2;
	ch->real_abils.intel -= 1;
	ch->real_abils.wis -= 1;
    break;
  case RACE_ORC_MORIHAI:
    ch->real_abils.intel = table[3];
	ch->real_abils.wis = table[4];
	ch->real_abils.con = table[1];
	ch->real_abils.str = table[0];
    ch->real_abils.dex = table[2];
	ch->real_abils.cha = table[5];
	ch->real_abils.cha -= 2;
    ch->real_abils.dex += 3;
    ch->real_abils.str += 1;
    break;
  case RACE_MYSTIC_CENTAUR:
    ch->real_abils.con = table[1];
    ch->real_abils.wis = table[4];
	ch->real_abils.str = table[0];
	ch->real_abils.intel = table[3];
    ch->real_abils.cha = table[5];
	ch->real_abils.dex = table[2];
	ch->real_abils.cha += 2;
    ch->real_abils.dex += 2;
    ch->real_abils.str += 1;
    break;
  case RACE_MYSTIC_MINOTAUR:
	ch->real_abils.str = table[0];
    ch->real_abils.con = table[1];
    ch->real_abils.wis = table[2];
	ch->real_abils.cha = table[3];
	ch->real_abils.intel = table[4];
	ch->real_abils.dex = table[5];
    ch->real_abils.dex += 1;
    ch->real_abils.con += 2;
	ch->real_abils.str += 2;
    break;
  case RACE_MYSTIC_AVIAN:
	ch->real_abils.dex = table[0];
	ch->real_abils.cha = table[1];
    ch->real_abils.con = table[2];
	ch->real_abils.intel = table[3];
    ch->real_abils.str = table[4];
    ch->real_abils.wis = table[5];
	ch->real_abils.cha += 1;
    ch->real_abils.dex += 3;
	ch->real_abils.wis += 2;
	ch->real_abils.str += 1;
    break;
  }  
  ch->real_abils.str_add = 0;

  ch->aff_abils = ch->real_abils;
}

/* Some initializations for characters, including initial skills */
void do_start(struct char_data *ch)
{
  int i;
	
  GET_LEVEL(ch) = 1;
  for (i = CLASS_APPRENTICE; i < NUM_CLASSES; i++)
    GET_CLASS_LEVEL(ch, i) = 0;
  GET_EXP(ch) = 1;

  set_title(ch, NULL);
  roll_real_abils(ch);

  GET_MAX_HIT(ch)  = 10;
  GET_MAX_MANA(ch) = 100;
  GET_MAX_MOVE(ch) = 82;

  switch (GET_CLASS(ch)) {

  case CLASS_MAGE:
    break;

  case CLASS_CLERIC:
    break;

  case CLASS_THIEF:
    SET_SKILL(ch, SKILL_SNEAK, 10);
    SET_SKILL(ch, SKILL_HIDE, 5);
    SET_SKILL(ch, SKILL_STEAL, 15);
    SET_SKILL(ch, SKILL_BACKSTAB, 10);
    SET_SKILL(ch, SKILL_PICK_LOCK, 10);
    SET_SKILL(ch, SKILL_TRACK, 10);
    break;

  case CLASS_SWORDFIGHTER:
    break;
  }

  advance_level(ch);
  mudlog(BRF, MAX(LVL_IMMORT, GET_INVIS_LEV(ch)), TRUE, "%s advanced to level %d", GET_NAME(ch), GET_LEVEL(ch));

  GET_HIT(ch) = GET_MAX_HIT(ch);
  GET_MANA(ch) = GET_MAX_MANA(ch);
  GET_MOVE(ch) = GET_MAX_MOVE(ch);

  GET_COND(ch, THIRST) = 24;
  GET_COND(ch, FULL) = 24;
  GET_COND(ch, DRUNK) = 0;

  if (CONFIG_SITEOK_ALL)
    SET_BIT_AR(PLR_FLAGS(ch), PLR_SITEOK);
}



/*
 * This function controls the change to maxmove, maxmana, and maxhp for
 * each class every time they gain a level.
 */
void advance_level(struct char_data *ch)
{
  int add_hp, add_mana = 0, add_move = 0, i;

  add_hp = con_app[GET_CON(ch)].hitp;

  switch (GET_CLASS(ch)) {

  case CLASS_APPRENTICE:
    add_hp += rand_number(10, 15);
    add_mana = rand_number(2, 4);
    add_move = rand_number(1, 3);
    break;

  case CLASS_SWORDFIGHTER:
    add_hp += rand_number(10, 15);
    add_mana = rand_number(2, 4);
    add_move = rand_number(1, 3);
    break;
  case CLASS_SWORDMASTER:
  case CLASS_WHIRLBLADE:
    add_hp += rand_number(15, 20);
    add_mana = rand_number(1, 3);
    add_move = rand_number(2, 4);
    break;
 
  case CLASS_MAGE:
    add_hp += rand_number(3, 8);
    add_mana = rand_number(GET_LEVEL(ch), (int)(1.5 * GET_LEVEL(ch)));
    add_mana = MIN(add_mana, 10);
    add_move = rand_number(0, 2);
    break;
  case CLASS_ARCHMAGE:
  case CLASS_WIZARD:
	add_hp += rand_number(4, 10);
    add_mana = rand_number(GET_LEVEL(ch), (int)(2.0 * GET_LEVEL(ch)));
    add_mana = MIN(add_mana, 15);
    add_move = rand_number(1, 3);
    break;

  case CLASS_RANGER:
	add_hp += rand_number(6, 11);
    add_mana = rand_number(3, 5);
    add_move = rand_number(3, 5);
    break;
  case CLASS_AVENGER:
  case CLASS_DRUID:
	add_hp += rand_number(11, 16);
    add_mana = rand_number(5, 7);
    add_move = rand_number(4, 8);
    break;

  case CLASS_CLERIC:
    add_hp += rand_number(5, 10);
    add_mana = rand_number(GET_LEVEL(ch), (int)(1.5 * GET_LEVEL(ch)));
    add_mana = MIN(add_mana, 10);
    add_move = rand_number(0, 2);
    break;
  case CLASS_MONK:
  case CLASS_PRIEST:
    add_hp += rand_number(6, 12);
    add_mana = rand_number(GET_LEVEL(ch), (int)(2.0 * GET_LEVEL(ch)));
    add_mana = MIN(add_mana, 15);
    add_move = rand_number(2, 4);
    break;

  case CLASS_THIEF:
    add_hp += rand_number(7, 13);
    add_mana = rand_number(0, 2);
    add_move = rand_number(1, 3);
    break;
  case CLASS_ROGUE:
  case CLASS_ASSASIN:
    add_hp += rand_number(10, 15);
    add_mana = rand_number(0, 3);
    add_move = rand_number(2, 4);
    break;

  case CLASS_CRUSADER:
    add_hp += rand_number(8, 14);
    add_mana = rand_number(GET_LEVEL(ch), (int)(1.3 * GET_LEVEL(ch)));
    add_mana = MIN(add_mana, 10);
    add_move = rand_number(1, 3);
    break;
  case CLASS_PALADIN:
  case CLASS_TEMPLAR:
    add_hp += rand_number(11, 16);
    add_mana = rand_number(GET_LEVEL(ch), (int)(1.7 * GET_LEVEL(ch)));
    add_mana = MIN(add_mana, 15);
    add_move = rand_number(2, 4);
    break;
  
  case CLASS_ELEMENTALIST:
    add_hp += rand_number(3, 8);
    add_mana = rand_number(GET_LEVEL(ch), (int)(1.5 * GET_LEVEL(ch)));
    add_mana = MIN(add_mana, 10);
    add_move = rand_number(0, 2);
    break;
  case CLASS_ILLUSIONIST:
  case CLASS_SORCERER:
	add_hp += rand_number(4, 10);
    add_mana = rand_number(GET_LEVEL(ch), (int)(2.0 * GET_LEVEL(ch)));
    add_mana = MIN(add_mana, 15);
    add_move = rand_number(1, 3);
    break;

  case CLASS_WARLOCK:
    add_hp += rand_number(8, 14);
    add_mana = rand_number(GET_LEVEL(ch), (int)(1.3 * GET_LEVEL(ch)));
    add_mana = MIN(add_mana, 10);
    add_move = rand_number(1, 3);
    break;
  case CLASS_LICH:
  case CLASS_DREADLORD:
    add_hp += rand_number(11, 16);
    add_mana = rand_number(GET_LEVEL(ch), (int)(1.7 * GET_LEVEL(ch)));
    add_mana = MIN(add_mana, 15);
    add_move = rand_number(2, 4);
    break;
  }

  ch->points.max_hit += MAX(1, add_hp);
  ch->points.max_move += MAX(1, add_move);

  if (GET_LEVEL(ch) > 1)
    ch->points.max_mana += add_mana;

  if (IS_MAGE(ch) || IS_CLERIC(ch))
    GET_PRACTICES(ch) += MAX(2, wis_app[GET_WIS(ch)].bonus);
  else
    GET_PRACTICES(ch) += MIN(2, MAX(1, wis_app[GET_WIS(ch)].bonus));

  if (GET_LEVEL(ch) >= LVL_IMMORT) {
    for (i = 0; i < 3; i++)
      GET_COND(ch, i) = (char) -1;
    SET_BIT_AR(PRF_FLAGS(ch), PRF_HOLYLIGHT);
  }

  snoop_check(ch);
  save_char(ch);
}


/*
 * This simply calculates the backstab multiplier based on a character's
 * level.  This used to be an array, but was changed to be a function so
 * that it would be easier to add more levels to your MUD.  This doesn't
 * really create a big performance hit because it's not used very often.
 */
int backstab_mult(int level)
{
  if (level <= 0)
    return 1;	  /* level 0 */
  else if (level <= 15)
    return 2;	  /* level 1 - 7 */
  else if (level <= 30)
    return 3;	  /* level 8 - 13 */
  else if (level <= 45)
    return 4;	  /* level 14 - 20 */
  else if (level <= 60)
    return 6;
  else if (level <= 75)
    return 7;
  else if (level <= 90)
    return 8;	  /* level 21 - 28 */
  else if (level < LVL_IMMORT)
    return 9;	  /* all remaining mortal levels */
  else
    return 20;	  /* immortals */
}


/*
 * invalid_class is used by handler.c to determine if a piece of equipment is
 * usable by a particular class, based on the ITEM_ANTI_{class} bitvectors.
 */
int invalid_class(struct char_data *ch, struct obj_data *obj)
{
  if (OBJ_FLAGGED(obj, ITEM_ANTI_MAGE) && IS_MAGE(ch))
    return TRUE;

  if (OBJ_FLAGGED(obj, ITEM_ANTI_CLERIC) && IS_CLERIC(ch))
    return TRUE;

  if (OBJ_FLAGGED(obj, ITEM_ANTI_RANGER) && IS_RANGER(ch))
    return TRUE;

  if (OBJ_FLAGGED(obj, ITEM_ANTI_SWORDFIGHTER) && IS_SWORDFIGHTER(ch))
    return TRUE;

  if (OBJ_FLAGGED(obj, ITEM_ANTI_CLERIC) && IS_CLERIC(ch))
    return TRUE;

  if (OBJ_FLAGGED(obj, ITEM_ANTI_CRUSADER) && IS_CRUSADER(ch))
    return TRUE;

  if (OBJ_FLAGGED(obj, ITEM_ANTI_THIEF) && IS_THIEF(ch))
    return TRUE;

  if (OBJ_FLAGGED(obj, ITEM_ANTI_ELEMENTALIST) && IS_ELEMENTALIST(ch))
    return TRUE;

  if (OBJ_FLAGGED(obj, ITEM_ANTI_WARLOCK) && IS_WARLOCK(ch))
    return TRUE;

  return FALSE;
}

/*
 * SPELLS AND SKILLS.  This area defines which spells are assigned to
 * which classes, and the minimum level the character must be to use
 * the spell or skill.
 */
void init_spell_levels(void)
{
  /* MAGES */
  spell_level(SPELL_MAGIC_MISSILE, CLASS_MAGE, 1);
  spell_level(SPELL_DETECT_INVIS, CLASS_MAGE, 2);
  spell_level(SPELL_DETECT_MAGIC, CLASS_MAGE, 2);
  spell_level(SPELL_CHILL_TOUCH, CLASS_MAGE, 3);
  spell_level(SPELL_INFRAVISION, CLASS_MAGE, 3);
  spell_level(SPELL_INVISIBLE, CLASS_MAGE, 4);
  spell_level(SPELL_ARMOR, CLASS_MAGE, 4);
  spell_level(SPELL_BURNING_HANDS, CLASS_MAGE, 5);
  spell_level(SPELL_LOCATE_OBJECT, CLASS_MAGE, 6);
  spell_level(SPELL_STRENGTH, CLASS_MAGE, 6);
  spell_level(SPELL_SHOCKING_GRASP, CLASS_MAGE, 7);
  spell_level(SPELL_SLEEP, CLASS_MAGE, 8);
  spell_level(SPELL_LIGHTNING_BOLT, CLASS_MAGE, 9);
  spell_level(SPELL_BLINDNESS, CLASS_MAGE, 9);
  spell_level(SPELL_DETECT_POISON, CLASS_MAGE, 10);
  spell_level(SPELL_COLOR_SPRAY, CLASS_MAGE, 11);
  spell_level(SPELL_ENERGY_DRAIN, CLASS_MAGE, 13);
  spell_level(SPELL_CURSE, CLASS_MAGE, 14);
  spell_level(SPELL_POISON, CLASS_MAGE, 14);
  spell_level(SPELL_FIREBALL, CLASS_MAGE, 15);
  spell_level(SPELL_CHARM, CLASS_MAGE, 16);
  spell_level(SPELL_ENCHANT_WEAPON, CLASS_MAGE, 26);
  spell_level(SPELL_CLONE, CLASS_MAGE, 30);

  /* CLERICS */
  spell_level(SPELL_CURE_LIGHT, CLASS_CLERIC, 1);
  spell_level(SPELL_ARMOR, CLASS_CLERIC, 1);
  spell_level(SPELL_CREATE_FOOD, CLASS_CLERIC, 2);
  spell_level(SPELL_CREATE_WATER, CLASS_CLERIC, 2);
  spell_level(SPELL_DETECT_POISON, CLASS_CLERIC, 3);
  spell_level(SPELL_DETECT_ALIGN, CLASS_CLERIC, 4);
  spell_level(SPELL_CURE_BLIND, CLASS_CLERIC, 4);
  spell_level(SPELL_BLESS, CLASS_CLERIC, 5);
  spell_level(SPELL_DETECT_INVIS, CLASS_CLERIC, 6);
  spell_level(SPELL_BLINDNESS, CLASS_CLERIC, 6);
  spell_level(SPELL_INFRAVISION, CLASS_CLERIC, 7);
  spell_level(SPELL_PROT_FROM_EVIL, CLASS_CLERIC, 8);
  spell_level(SPELL_POISON, CLASS_CLERIC, 8);
  spell_level(SPELL_GROUP_ARMOR, CLASS_CLERIC, 9);
  spell_level(SPELL_CURE_CRITIC, CLASS_CLERIC, 9);
  spell_level(SPELL_SUMMON, CLASS_CLERIC, 10);
  spell_level(SPELL_REMOVE_POISON, CLASS_CLERIC, 10);
  spell_level(SPELL_WORD_OF_RECALL, CLASS_CLERIC, 12);
  spell_level(SPELL_EARTHQUAKE, CLASS_CLERIC, 12);
  spell_level(SPELL_DISPEL_EVIL, CLASS_CLERIC, 14);
  spell_level(SPELL_DISPEL_GOOD, CLASS_CLERIC, 14);
  spell_level(SPELL_SANCTUARY, CLASS_CLERIC, 15);
  spell_level(SPELL_CALL_LIGHTNING, CLASS_CLERIC, 15);
  spell_level(SPELL_HEAL, CLASS_CLERIC, 16);
  spell_level(SPELL_CONTROL_WEATHER, CLASS_CLERIC, 17);
  spell_level(SPELL_SENSE_LIFE, CLASS_CLERIC, 18);
  spell_level(SPELL_HARM, CLASS_CLERIC, 19);
  spell_level(SPELL_GROUP_HEAL, CLASS_CLERIC, 22);
  spell_level(SPELL_REMOVE_CURSE, CLASS_CLERIC, 26);

  /* THIEVES */
  spell_level(SKILL_SNEAK, CLASS_THIEF, 1);
  spell_level(SKILL_PICK_LOCK, CLASS_THIEF, 2);
  spell_level(SKILL_BACKSTAB, CLASS_THIEF, 3);
  spell_level(SKILL_STEAL, CLASS_THIEF, 4);
  spell_level(SKILL_HIDE, CLASS_THIEF, 5);
  spell_level(SKILL_TRACK, CLASS_THIEF, 6);

  /* WARRIORS */
  spell_level(SKILL_KICK, CLASS_SWORDFIGHTER, 1);
  spell_level(SKILL_RESCUE, CLASS_SWORDFIGHTER, 3);
  spell_level(SKILL_TRACK, CLASS_SWORDFIGHTER, 9);
  spell_level(SKILL_BASH, CLASS_SWORDFIGHTER, 12);
}


/*
 * This is the exp given to implementors -- it must always be greater
 * than the exp required for immortality, plus at least 20,000 or so.
 */
#define EXP_MAX  500000000

/* Function to return the exp required for each class/level */
int level_exp(int chclass, int level, int race)
{
  int modifier;
  if (level > LVL_IMPL || level < 0) {
    log("SYSERR: Requesting exp for invalid level %d!", level);
    return 0;
  }
  if(level < 2)
    return (level);

  if (level > LVL_IMMORT)
    return (EXP_MAX - ((LVL_IMPL - level) * 1000));

  /* Exp required for normal mortals is below */
  switch (chclass) {
  case CLASS_APPRENTICE:
	/* See further in code. May not gain exp. */
	modifier = 2500;
	break;
  
  case CLASS_SWORDFIGHTER:
  case CLASS_SWORDMASTER:
  case CLASS_WHIRLBLADE:
    modifier = 1500;
    break;
  case CLASS_MAGE:
  case CLASS_ARCHMAGE:
  case CLASS_WIZARD:
    modifier = 2000;
    break;
  case CLASS_RANGER:
  case CLASS_AVENGER:
  case CLASS_DRUID:
	modifier = 1700;
	break;
  case CLASS_CLERIC:
  case CLASS_MONK:
  case CLASS_PRIEST:
    modifier = 1500;
    break;
  case CLASS_THIEF:
  case CLASS_ROGUE:
  case CLASS_ASSASIN:
    modifier = 1000;
    break;
  case CLASS_CRUSADER:
  case CLASS_PALADIN:
  case CLASS_TEMPLAR:
	modifier = 2000;
	break;
  case CLASS_ELEMENTALIST:
  case CLASS_ILLUSIONIST:
  case CLASS_SORCERER:
	modifier = 2000;
	break;
  case CLASS_WARLOCK:
  case CLASS_LICH:
  case CLASS_DREADLORD:
	modifier = 2500;
	break;

  default:
    log("SYSERR: Requesting exp for invalid class at level_exp().");
    return 0;
  }

  switch(race) {
  case RACE_HOBBIT_HARFOOT:
  case RACE_HOBBIT_STOOR:
  case RACE_HOBBIT_FALLOHIDE:
    modifier *= 0.90;
    break;
  case RACE_ELF_VANYAR:
  case RACE_ELF_NOLDOR:
	modifier *= 1.20;
    break;
  case RACE_ELF_TELERI:
    /* modifier = 1 */
    break;
  case RACE_HUMAN_AIEL:
	modifier *= 1.25;
    break;
  case RACE_HUMAN_NUMERON:
  case RACE_HUMAN_HILDOR:
	modifier *= 1.10;
    break;
  case RACE_HUMAN_ENGWAR:
	modifier *= 1.05;
    break;
  case RACE_ORC_URUKHAI:
	modifier *= 1.15;
    break;
  case RACE_ORC_OLEGHAI:
	modifier *= 1.15;
    break;
  case RACE_ORC_MORIHAI:
	modifier *= 1.20;
    break;
  case RACE_MYSTIC_CENTAUR:
	modifier *= 1.50;
    break;
  case RACE_MYSTIC_MINOTAUR:
	modifier *= 1.50;
    break;
  case RACE_MYSTIC_AVIAN:
	modifier *= 1.50;
    break;
  default:
    log("SYSERR: Requesting exp for invalid race at level_exp().");
    return 0;
  }
  
  modifier *= pow(1.11, (level - 2)); 
  modifier += 1000 * level;
  return (modifier);
}







